import type { APIRoute } from 'astro'
import type APIRoutes from 'config/api-routes'
import { formatRoute } from 'route'
import StatusCode from './http/status-code'
import ResponseBuilder from './response-builder'

// Typescript is dumb so we need to add this `THANKS_TS` interface to make the different runtime typing works
interface THANKS_TS<Output = any, Input = never> {
	_input?: Input
	_output?: Output
}

// create the overload of APIRoute to add type safety
export type TypeSafeAPIRoute<Output = any, Input = never> = APIRoute & THANKS_TS<Output, Input>

// extract both the input object and output object
export type OutputParam<T> = T extends THANKS_TS<infer U, any> ? U : any
export type InputParam<T> = T extends THANKS_TS<any, infer U> ? U : any

// type TestO = OutputParam<TypeSafeAPIRoute<string, number>
// type TestI = InputParam<TypeSafeAPIRoute<string, number>
// type TestN1 = InputParam<APIRoute>
// type TestN2 = InputParam<APIRoute>

// extract the URL parameters
type ExtractParameters<T> = T extends `${string}[${infer Param}]${string}`
	? Param | ExtractParameters<Exclude<T, T>>
	: never

/**
 * make a type safe API request
 *
 * @param path the path ro request (exclude the `/api` part)
 * @param opts the options to user
 * @param opts.params the parameters to input to the original url
 * @param opts.method the method to request
 * @param opts.body the body of the request, it can be a raw object to be sent as JSON or a `FormData` object
 * @returns the response from the API endpoint
 * @throws an error if statusCode is above or equal to 400
 */
export default async function api<R extends keyof APIRoutes, M extends keyof APIRoutes[R] = keyof APIRoutes[R]>(
	path: R,
	opts?: {
		params?: Record<ExtractParameters<R>, string | number>
		method?: M
		body?: InputParam<APIRoutes[R][M]>
	}
): Promise<OutputParam<APIRoutes[R][M]>> {
	const { params, method = 'GET', body } = opts ?? {}

	// fetch the remote
	const res = await fetch('/api' + formatRoute(path, params), {
		method: method as string,
		body: typeof body === 'undefined' || body instanceof FormData ? body ?? null : JSON.stringify(body)
	})

	// throw if an error occur
	if (res.status >= 400) {
		throw new Error(`${path as string} status code invalid :(`)
	}

	return res.json() as OutputParam<APIRoutes[R][M]>
}

type PromiseLike<T> = Promise<T> | T

/**
* make a type safe API Route
* @param baseFunction the original APIRoute function with the additionnal possible return of an object
* @returns the enhanced `APIRoute` function
*/
export function makeAPI<Output = any, Input = never>(
	baseFunction: (...params: Parameters<APIRoute>) => PromiseLike<Response | { body: Output, status: number } | Output>
): TypeSafeAPIRoute<Output, Input> {
	return async (ctx) => {
		const res = await baseFunction(ctx)
		if (res instanceof Response) {
			return res
		}
		if (typeof res === 'object' && res !== null && 'body' in res && 'status' in res && Object.keys(res).length === 2) {
			return new ResponseBuilder()
				.body(res.body as object)
				.status(res.status)
				.build()
		}
		return new ResponseBuilder()
			.body(res as object)
			.status(StatusCode.OK)
			.build()
	}
}
