import type { Console } from 'node:console'
import type LoggerInterface from './LoggerInterface'

/**
 * Special class that is meant to replace `global.console` with it to be able to handle other logs
 */
export default class ConsoleReplacer implements Console {
	public readonly original: Console

	/** @ts-expect-error normal behavior */
	public Console: console.ConsoleConstructor = null

	public constructor(
		private readonly logger: LoggerInterface
	) {
		let original = console
		while (original instanceof ConsoleReplacer) {
			original = original.original
		}
		this.original = original
	}

	public assert(_condition?: boolean, ..._data: Array<any>): void {
		throw new Error('Method not implemented.')
	}
	public clear(): void {/** normal */ }

	public count(_label?: string): void {
		throw new Error('Method not implemented.')
	}
	public countReset(_label?: string): void {
		throw new Error('Method not implemented.')
	}
	public debug(...data: Array<any>): void {
		this.logger.debug(this.parseData(data))
	}
	public dir(_item?: any, _options?: any): void {
		throw new Error('Method not implemented.')
	}
	public dirxml(..._data: Array<any>): void {
		throw new Error('Method not implemented.')
	}
	public error(...data: Array<any>): void {
		this.logger.error(this.parseData(data))
	}
	public group(..._data: Array<any>): void {
		throw new Error('Method not implemented.')
	}
	public groupCollapsed(..._data: Array<any>): void {
		throw new Error('Method not implemented.')
	}
	public groupEnd(): void
	public groupEnd(): void
	public groupEnd(): void {
		throw new Error('Method not implemented.')
	}
	public info(...data: Array<any>): void {
		this.logger.info(this.parseData(data))
	}
	public log(...data: Array<any>): void {
		this.logger.info(this.parseData(data))
	}
	public table(_tabularData?: any, _properties?: Array<string>): void {
		throw new Error('Method not implemented.')
	}
	public time(_label?: string): void {
		throw new Error('Method not implemented.')
	}
	public timeEnd(label?: string): void
	public timeEnd(label?: string): void
	public timeEnd(_label?: unknown): void {
		throw new Error('Method not implemented.')
	}
	public timeLog(_label?: string, ..._data: Array<any>): void {
		throw new Error('Method not implemented.')
	}
	public timeStamp(_label?: string): void {
		throw new Error('Method not implemented.')
	}
	public trace(...data: Array<any>): void {
		const trace: { stack?: string } = {}
		// normal behavior, this.trace is used as the base as so to not include it's call in the trace
		// eslint-disable-next-line @typescript-eslint/unbound-method
		Error.captureStackTrace(trace, this.trace)
		this.logger.debug(`${this.parseData(data)} ${trace.stack}`)
	}
	public warn(...data: Array<any>): void {
		this.logger.warning(this.parseData(data))
	}
	public profile(_label?: string): void {
		throw new Error('Method not implemented.')
	}
	public profileEnd(_label?: string): void {
		throw new Error('Method not implemented.')
	}

	private parseData(data: Array<any>): string {
		return data.map((it) => {
			if (it === null) {
				return 'null'
			}
			if (typeof it === 'string') {
				return it
			}
			let strVersion = '[object Object]'
			if (it instanceof Error) {
				strVersion = it.toString()
			}
			if (typeof it === 'object' && 'toString' in it) {
				// eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-member-access
				strVersion = it.toString()
			}
			if (strVersion !== '[object Object]') {
				return strVersion
			}
			return JSON.stringify(it)
		}).join(' ')
	}
}
