import type { SchemaInfer, SchemaObject } from '@dzeio/schema'
import type { Query } from 'models/query'

type OneOrMany<T> = T | Array<T>

export interface DBPull<M extends SchemaObject> {
	/**
	 * total number of rows that are valid with the specified query
	 */
	rows: number
	/**
	 * total number of rows in the table
	 */
	rowsTotal: number

	/**
	 * current page number
	 */
	page: number

	/**
	 * total amount of pages
	 */
	pageTotal: number

	/**
	 * the data fetched
	 */
	data: Array<SchemaInfer<M>>
}

/**
 * the Dao is the object that connect the Database or source to the application layer
 *
 * you MUST call it through the `DaoFactory` file
 */
export default interface DaoAdapter<M extends SchemaObject> {
	/**
	 * create a new object in the remote source
	 *
	 * @param obj the object to create
	 */
	create?(obj: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>
	create?(obj: Array<Partial<SchemaInfer<M>>>): Promise<Array<SchemaInfer<M> | null>>
	create?(obj: OneOrMany<Partial<SchemaInfer<M>>>): Promise<OneOrMany<SchemaInfer<M> | null>>

	/**
	 * read from the remote source
	 *
	 * @param query the query to filter/sort results
	 */
	read?(query?: Query<SchemaInfer<M>>): Promise<DBPull<M>>

	/**
	 * update an object to the remote source
	 *
	 * @param obj the object to update
	 */
	update?(obj: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>

	/**
	 * depending if the object already exists or not
	 * it will update an exisintg object or create a new one
	 *
	 * @param obj the object to insert/update
	 */
	upsert?(obj: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>

	/**
	 * update an object to the remote source
	 *
	 * @param id (DEPRECATED) the ID of the object
	 * @param obj the object to patch (MUST include ids, and changes)
	 */
	patch?(id: string, obj: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>

	/**
	 * delete an object from the source
	 * @param obj the object ot delete (it must at least include the id(s))
	 */
	delete?(obj: Partial<SchemaInfer<M>>): Promise<boolean>
}
