/* eslint-disable @typescript-eslint/no-non-null-assertion */
import { objectFind, objectKeys, objectLoop, objectMap, objectRemap, objectSize, objectValues } from '@dzeio/object-util'

import type { SchemaInfer, SchemaItem, SchemaObject } from '@dzeio/schema'
import logger from 'config/logger'
import PostgresClient from 'models/clients/postgres-client'
import { Sort, type Query } from 'models/query'
import crypto from 'node:crypto'
import type DaoAdapter from '.'
import type { DBPull } from '.'
import { filter } from './adapter-utils'

type OneOrMany<T> = T | Array<T>

const specialKeywords = ['user', 'end']
type DBValue = string | number | boolean | Date

export default class PostgresAdapter<M extends SchemaObject> implements DaoAdapter<M> {
	private readonly id: Array<string> = []

	public constructor(
		/**
		 * the schema used by Cassandra
		 */
		public readonly schema: M,
		/**
		 * the table name
		 */
		public readonly table: string,

		/**
		 * additionnal options to make the adapter work
		 */
		private readonly options?: {
			/**
			 * log the requests made to cassandra
			 */
			debug?: boolean
		}
	) {
		objectLoop(this.schema.model as Record<string, SchemaItem>, (tmp, key) => {
			if (tmp.attributes.includes('db:auto')) {
				this.id.push(key)
			}
		})
	}

	// TODO: make it clearer what it does
	public async create(obj: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>
	public async create(obj: Array<Partial<SchemaInfer<M>>>): Promise<Array<SchemaInfer<M> | null>>
	public async create(original: OneOrMany<Partial<SchemaInfer<M>>>): Promise<OneOrMany<SchemaInfer<M> | null>> {
		// Convert to an array of elements
		const obj = Array.isArray(original) ? original : [original]
		if (!obj.length) {
			return []
		}

		// handle automated values
		objectLoop(this.schema.model as Record<string, SchemaItem>, (item, key) => {
			if (item.attributes.includes('db:created') || item.attributes.includes('db:updated')) {
				// @ts-expect-error things get validated anyway
				obj.forEach((it) => it[key] = new Date())
			} else if (item.attributes.includes('db:auto')) {
				obj.forEach((objet) => {
					if (objet[key]) {
						return
					}

					if (item.isOfType('')) {
						// @ts-expect-error things get validated anyway
						objet[key] = crypto.randomBytes(16).toString('hex')
					} else if (item.isOfType(123)) {
						// @ts-expect-error things get validated anyway
						objet[key] = crypto.randomBytes(16).readUint32BE()
					} else {
						throw new Error('cannot generate ID because it is not compatible with it')
					}
				})
			}
		})

		// parse the data with the Model
		const clone = obj.map((o) => {
			const { object, errors } = this.schema.parse(o)
			if (errors) {
				console.error(errors)
				throw new Error(`Invalid data given to create the final object: ${JSON.stringify(errors)}`)
			}
			return object
		})

		// prepare the database query
		const keys = objectKeys(clone.reduce((a, it) => ({ ...a, ...it }), {}))
			.map((it) => {
				if (typeof it === 'string' && specialKeywords.includes(it)) { // handle the special keyword
					return `"${it}"`
				}
				return it
			})
		const keysStr = keys.join(', ')
		const valuesStr = clone.map((_, i) => {
			const keyValuesStr = keys.map((_1, idx) => `$${i * keys.length + idx + 1}`).join(', ')
			return `(${keyValuesStr})`
		})
		const req = `INSERT INTO ${this.table} (${keysStr}) VALUES ${valuesStr.join(', ')};`
		const client = await PostgresClient.get()

		// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
		const params = clone.flatMap((o: any) => keys.map((key) => this.valueToDB(key, o[key])))

		if (this.options?.debug) {
			logger.debug(req, params)
		}

		// send to the database
		try {
			await client.execute(req, params)
		} catch (e) {
			logger.debug(e, req, params)
			return null
		}
		const transform = (o: SchemaObject) => this.schema.parse(o).object ?? null
		// @ts-expect-error conditionnal output based on original input
		return (Array.isArray(original) ? clone.map(transform) : transform(clone[0]))
	}

	// eslint-disable-next-line complexity
	public async read(query?: Query<SchemaInfer<M>>): Promise<DBPull<M>> {
		// prepare the request to the database based on the query parameters
		const req: Array<string> = ['SELECT', '*', 'FROM', this.table]
		const params: Array<DBValue> = []
		if (query) {
			const keyValues: Record<string, string | number | boolean | Date> = {}
			objectLoop(query, (value, key) => {
				if (typeof value === 'object' || String(key).startsWith('$')) {
					return
				}
				keyValues[String(key)] = this.valueToDB(key, value as any)
			})
			if (objectSize(keyValues)) {
				req.push('WHERE')
				objectLoop(keyValues, (value, key, index) => {
					params.push(value)
					req.push(`"${key}"=$${index + 1}`)
					if (index < objectSize(keyValues) - 1) {
						req.push('AND')
					}
				})
			}
		}

		const client = await PostgresClient.get()

		if (this.options?.debug) {
			logger.debug(req)
		}

		// read from the database
		let res: Array<Record<string, any>>
		try {
			res = await client.execute(req.join(' '), params)
		} catch (error) {
			console.error('error running request')
			console.error(req)
			throw error
		}
		if (res.length === 0) {
			return {
				rows: 0,
				pageTotal: 0,
				page: 1,
				rowsTotal: 0,
				data: []
			}
		}

		if (this.options?.debug) {
			logger.debug('preEdits', res)
		}

		// post-process the data from the database
		const raw = res
			.map((obj) => {
				// remap to use system value instead of db values
				obj = objectRemap(this.schema.model as Record<string, SchemaItem>, (_, key) => ({
					key,
					value: this.dbToValue(key as any, (obj as Record<string, unknown>)[key])
				}))

				// validate the schema
				const tmp = this.schema.parse(obj)
				if (tmp.valid) {
					return tmp.object
				}
				logger.debug(tmp.errors)
				return null
			})
			.filter((it): it is SchemaInfer<M> => !!it)

		// temp modification of comportement to use the new and better query system
		if ((!query?.$sort) && objectFind(this.schema.model as Record<string, SchemaItem>, (_, key) => key === 'created')) {
			// temp fix for the sorting algorithm
			if (!query) {
				// @ts-expect-error normal currently
				query = { $sort: { created: Sort.DESC } }
			} else {
				query.$sort = { created: Sort.DESC }
			}
		}
		let dataset = raw


		if (this.options?.debug) {
			logger.debug('preFilters', dataset)
		}

		if (query) {
			dataset = filter(query, dataset, this.options).filtered
		}
		return {
			rows: dataset.length,
			rowsTotal: res.length,
			page: 1,
			pageTotal: 1,
			// page: page,
			// pageTotal: pageLimit ? res.rowLength / pageLimit : 1,
			data: dataset
		}
	}

	public async update(obj: SchemaInfer<M>): Promise<SchemaInfer<M> | null> {
		return this.patch(obj)
	}

	public async patch(id: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>
	public async patch(id: string, obj: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null>
	// eslint-disable-next-line complexity
	public async patch(id: string | Partial<SchemaInfer<M>>, obj?: Partial<SchemaInfer<M>>): Promise<SchemaInfer<M> | null> {
		if (!obj) {
			if (typeof id === 'string') {
				return null
			}
			obj = { ...id }
		}

		// const tmp = this.schema.validate(obj)
		// // if (tmp.error) {
		// // 	throw new Error(`obj invalid can\'t patch ${JSON.stringify(tmp.error)}`)
		// // }

		// obj = tmp.object

		// update the updated time
		objectLoop(this.schema.model as Record<string, SchemaItem>, (item, key) => {
			if (item.attributes.includes('db:updated')) {
				// @ts-expect-error things get validated anyway
				obj[key] = new Date()
			}
		})

		// build the request parts
		const parts: Array<string> = ['UPDATE', this.table, 'SET']
		const params: Array<any> = []

		// remove ids
		for (const tmp of this.id) {
			// eslint-disable-next-line @typescript-eslint/no-dynamic-delete
			delete obj[tmp]
		}

		// map the items to update
		const keys = objectMap(obj as object, (_, key: string, idx) => {
			if (specialKeywords.includes(key)) {
				return `"${key}"=$${idx + 1}`
			}

			return `${key}=$${idx + 1}`
		})
		parts.push(keys.join(', '))
		params.push(...objectValues(obj as Record<string, unknown>))

		// filter by the ids
		parts.push('WHERE')
		const read: Partial<any> = {}
		for (let idx = 0; idx < this.id.length; idx++) {
			const key = this.id[idx]!

			if (idx > 0) {
				parts.push('AND')
			}
			parts.push(`${key}=$${params.length + 1}`)
			const value = obj[key] ?? (typeof id === 'string' ? id : id[key]) as SchemaInfer<M>[keyof M]
			read[key] = this.valueToDB(key, value)
			if (!value) {
				throw new Error(`Missing id (${key})`)
			}
			params.push(value)
		}

		const req = parts.join(' ')
		const client = await PostgresClient.get()

		if (this.options?.debug) {
			logger.debug(req, params)
		}

		try {
			const res = await client.execute(req, params)
			// logger.debug(res, req)
			if (this.options?.debug) {
				logger.debug('post patch result', res, req)
			}
			return (await this.read(read)).data[0] ?? null
		} catch (e) {
			logger.debug(e, req, params)
		}
		return null
	}

	public async delete(obj: Partial<SchemaInfer<M>>): Promise<boolean> {
		const parts = ['DELETE', 'FROM', this.table, 'WHERE']

		objectLoop(obj as SchemaInfer<M>, (value, key: string, idx) => {
			if (idx > 0) {
				parts.push('AND')
			}

			if (specialKeywords.includes(key)) {
				key = `"${key}"`
			}
			switch (typeof value) {
				case 'string':
					parts.push(`${key} = '${value}'`)
					break

				default:
					parts.push(`${key} = ${value as any}`)
					break
			}
		})

		const client = await PostgresClient.get()

		try {
			await client.execute(parts.join(' '))
		} catch (e) {
			console.error(e, parts)
			throw e
		}
		return true
	}

	private valueToDB(key: keyof SchemaInfer<M>, value: unknown): DBValue {
		const item: SchemaItem<unknown> = (this.schema.model as Record<typeof key, SchemaItem>)[key]

		if (item.isOfType({})) {
			return JSON.stringify(value)
		}

		return value as DBValue
	}

	private dbToValue<Key extends keyof SchemaInfer<M>>(key: Key, value: unknown): SchemaInfer<M>[Key] {
		const item: SchemaItem<unknown> = (this.schema.model as Record<Key, SchemaItem>)[key]

		if (value instanceof Date) {
			// tranform the Date back into an UTC one
			const localeDateString = value.toISOString()
			return new Date(localeDateString) as SchemaInfer<M>[Key]
		}

		if (item.isOfType(543) && typeof value === 'string') {
			return parseFloat(value) as SchemaInfer<M>[Key]
		}

		if (item.isOfType({}) && typeof value === 'string') {
			return JSON.parse(value) as SchemaInfer<M>[Key]
		}

		return value as SchemaInfer<M>[Key]
	}
}
