import config from 'config/models'
import type MigrationObj from 'models/migrations'

export enum ConnectionStatus {
	DISCONNECTED,
	MIGRATING,
	READY
}

export interface ClientStatic<C extends Client = Client> {
	get(): Promise<C>
}

export default abstract class Client {

	public status: ConnectionStatus = ConnectionStatus.DISCONNECTED

	/**
	 * -1: unknown
	 * 0: migrating
	 * 1: migrated
	 */
	public isMigrating = false

	/**
	 * Migrate the database to the latest version
	 */
	public async migrateToLatest() {
		if (await this.isMigrated() || this.isMigrating) {
			return
		}
		this.isMigrating = true

		const migrations = this.getLocalMigrations()
		const remotes = await this.loadMigrations()

		const missingFromRemote = migrations.filter((it) => !remotes.includes(it.date))
		for (const migration of missingFromRemote) {
			await migration.up(this)
			await this.setMigrated(migration)
		}

		this.isMigrating = false
	}

	public getLocalMigrations(): ReadonlyArray<MigrationObj> {
		return config.migrations as ReadonlyArray<MigrationObj>
	}

	/**
	 * wait until every migrations are done or fail
	 */
	public async waitForMigrations(): Promise<void> {
		// skip if already migrated
		if (await this.isMigrated()) {
			return
		}

		// start migrations if not started
		if (!this.isMigrating) {
			await this.migrateToLatest()
		}

		// wait for migrations to end
		while (!await this.isMigrated()) {
			await new Promise((res) => {
				setTimeout(res, 100)
			})
		}
	}

	public async isMigrated(): Promise<boolean> {
		const migrations = this.getLocalMigrations()
		const remotes = await this.loadMigrations()

		// TODO: compare migrations to remotes
		return migrations.length === remotes.length
	}

	/**
	 * load migrations that are in the system
	 */
	public abstract loadMigrations(): Promise<Array<number>>

	/**
	* indicate that the migrations was finished
	*/
	public abstract setMigrated(migration: MigrationObj): Promise<void>

	public abstract execute(query: string, params?: Array<unknown> | object, ...options: Array<any>): Promise<Array<Record<string, unknown>>>

	public abstract connect(): Promise<void>

	/**
	 * indicate if the client is ready for new requests (not if migrations are done or not)
	 */
	public abstract isReady(): Promise<boolean>
}
