import Logger from 'config/logger'
import { getEnv, requireEnv } from 'libs/env'
import type Migration from 'models/migrations'
import pg from 'pg'
import Client from '.'
const Postgres = pg.Pool

function parseIntUndefined(input: string | undefined) {
	if (typeof input === 'string') {
		return Number.parseInt(input, 10)
	}
	return input
}

interface Request {
	query: string
	params: Array<unknown> | object | undefined
}

export default class PostgresClient extends Client {
	private static instance: PostgresClient = new PostgresClient()
	private pool?: pg.Pool | null
	private lastRequest: Request | undefined

	/**
	 * get the connexion to cassandra, it will try until it succedeed
	 */
	public static async get() {
		return PostgresClient.instance
	}

	public override async loadMigrations(): Promise<Array<number>> {
		try {
			const res = (await this.execute('SELECT migration FROM migrations'))
				.map((it) => typeof it.migration === 'number' ? it.migration : Number.parseInt(it.migration as string, 10))
			return res
		} catch (err) {
			// table does not exists
			await this.execute('CREATE TABLE IF NOT EXISTS migrations (migration INT PRIMARY KEY, created TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP );')
			Logger.alert('Settings table does not exists')
			Logger.alert(err)
			return []
		}
	}
	public override async setMigrated(version: Migration): Promise<void> {
		await this.execute('INSERT INTO migrations (id) VALUES ($1);', [version.date.toString()])
	}
	public override async execute(query: string, params?: Array<unknown>, ..._options: Array<unknown>): Promise<Array<Record<string, unknown>>> {
		if (!this.pool || !await this.isReady()) {
			throw new Error('not connected')
		}
		this.lastRequest = { query, params }

		const client = await this.pool.connect()
		try {
			// Logger.debug('db', query, params)
			const res = await client.query<Record<string, unknown>>(query, params)

			// eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
			if (!res.rows) {
				throw new Error('Strangely, Postgres does not return the `rows` parameter :(')
			}

			return res.rows
		} catch (err) {
			console.error(err)
			throw new Error('Error connecting to Postgres')
		} finally {
			client.release()
		}
	}

	public override async connect(): Promise<void> {
		if (this.pool) {
			return
		}
		Logger.info('starting Postgres client')

		this.pool = new Postgres({
			host: requireEnv('POSTGRES_HOST'),
			user: requireEnv('POSTGRES_USERNAME'),
			password: requireEnv('POSTGRES_PASSWORD'),
			port: Number.parseInt(getEnv('POSTGRES_PORT', '5432'), 10),
			database: requireEnv('POSTGRES_DATABASE', 'projectmanager'),
			connectionTimeoutMillis: parseIntUndefined(getEnv('POSTGRES_CONNECTION_TIMEOUT')),
			idleTimeoutMillis: parseIntUndefined(getEnv('POSTGRES_IDLE_TIMEOUT')),
			max: parseIntUndefined(getEnv('POSTGRES_MAX_CLIENT')),
			// debug(connection, query, parameters, paramTypes) {
			// 		console.log(`${query}, ${parameters}`);
			// },
		})
			.on('error', (err) => {
				console.error('AN ERROR OCCURED INSIDE POSTGRESQL', err)
				console.error('THINGS MIGHT NOT WORK UNTIL RESTARTED')
				console.error('RESTARTING THE POSTGRESQL CLIENT TO TRY AGAIN.')
				if (this.lastRequest) {
					console.error('LAST REQUEST:', this.lastRequest.query, this.lastRequest.params)
				}
				this.pool = null
				void this.connect()
			})
		Logger.info('Connected to postgres')
	}

	public override async isReady(): Promise<boolean> {
		return !!this.pool
	}
}
