import { objectPick, objectRemap } from '@dzeio/object-util'
import config from 'config/models'
import type Client from 'models/clients'
import type Dao from 'models/dao'
import type Migration from 'models/migrations'

export interface DaoConfig {
	/**
	 * the main client used to handle migrations from
	 */
	mainClient: Client | null
	/**
	 * list the migrations of the app
	 */
	migrations: Array<Migration> | null
	/**
	 * indicate the list of models used through the app
	 */
	models: Record<string, Dao>
}

/**
 * Class to get any DAO
 */

type Remapped<T> = {
	// @ts-expect-error it works so idc
	[K in keyof T as `${K}Dao`]: T[K]
}

// biome-ignore lint/complexity/noStaticOnlyClass: <explanation>
export default class DaoFactory {
	/**
	 * get the total list of daos available
	 * @returns return the list of daos available
	 */
	public static getAll(): Record<string, Dao> {
		return config.models
	}

	/**
	 * Get a a dao by its key
	 *
	 * it will throw an error if no Dao exists linked to the item key
	 *
	 * @param key the dao key to get
	 * @returns the Dao you want as a singleton
	 */
	public static get<Key extends keyof typeof config['models']>(...key: [Key]): typeof config['models'][Key]
	public static get<Key extends keyof typeof config['models']>(...key: Array<Key>): Remapped<Pick<typeof config['models'], Key>>
	public static get<Key extends keyof typeof config['models']>(...key: Array<Key>): Remapped<Pick<typeof config['models'], Key>> {
		if (key.length === 1) {
			// @ts-expect-error normal behavior
			return config.models[key[0]]
		}
		const t = objectPick(config.models, ...key)
		return objectRemap(t, (v, k) => ({
			// @ts-expect-error normal behavior
			key: `${k}Dao`,
			value: v
		}))
	}

	/**
	 * get the main client linked to migrations
	 * @returns the main client
	 */
	public static async client(): ReturnType<(typeof config.mainClient)['get']> {
		return config.mainClient.get()
	}
}
