import type { SchemaInfer, SchemaObject } from '@dzeio/schema'
import Logger from 'config/logger'
import type DaoAdapter from './adapters'

/**
 * the Dao is the object that connect the Database or source to the application layer
 *
 * you MUST call it through the `DaoFactory` file test commit
 */
export default class Dao<S extends SchemaObject = SchemaObject> {

	public constructor(
		public readonly schema: S,
		public readonly adapter: DaoAdapter<S>
	) { }

	/**
	 * insert a new object into the source
	 *
	 * @param obj the object to create
	 * @returns the object with it's id filled if create or null otherwise
	 */
	public async create(...params: Parameters<NonNullable<DaoAdapter<S>['create']>>): ReturnType<NonNullable<DaoAdapter<S>['create']>> {
		if (!this.adapter.create) {
			throw new Error('the Adapter does not allow you to create elements')
		}
		return this.adapter.create(...params)
	}

	/**
	 * insert a new object into the source
	 *
	 * @param obj the object to create
	 * @returns the object with it's id filled if create or null otherwise
	 */
	public async insert(...params: Parameters<NonNullable<DaoAdapter<S>['create']>>): ReturnType<NonNullable<DaoAdapter<S>['create']>> {
		return this.create(...params)
	}

	/**
	 * find the list of objects having elements from the query
	 *
	 * @param query a partial object which filter depending on the elements, if not set it will fetch everything
	 * @returns an array containing the list of elements that match with the query
	 */

	public async findAll(...params: Parameters<NonNullable<DaoAdapter<S>['read']>>): ReturnType<NonNullable<DaoAdapter<S>['read']>> {
		if (!this.adapter.read) {
			throw new Error('the Adapter does not allow you to read from the remote source')
		}

		const req = this.adapter.read(...params)
		try {
			return await Promise.race([req, new Promise<void>((_, rej) => {
				setTimeout(() => {
					rej(new Error('timeout running request'))
				}, 3000)
			})]) as Awaited<typeof req>
		} catch (err) {
			Logger.error('DAO error:', err)
			return {
				rows: 0,
				rowsTotal: 0,
				page: 1,
				pageTotal: 0,
				data: []
			}
		}
	}

	/**
	 * find the list of objects having elements from the query
	 *
	 * @param query a partial object which filter depending on the elements, if not set it will fetch everything
	 * @returns an array containing the list of elements that match with the query
	 */
	public async find(...params: Parameters<NonNullable<DaoAdapter<S>['read']>>): ReturnType<NonNullable<DaoAdapter<S>['read']>> {
		return this.findAll(...params)
	}

	/**
	 * find an object by it's id
	 *
	 * (shortcut to findOne({id: id}))
	 *
	 * @param id the id of the object
	 * @returns
	 */
	public findById(id: any) {
		// eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
		return this.findOne({ id: id })
	}

	/**
	 * find an object by it's id
	 *
	 * (shortcut to `findOne({ id: id })`)
	 *
	 * @param id the id of the object
	 * @returns
	 */
	public get(id: any) {
		return this.findById(id)
	}

	/**
	 * find the first element that match `query`
	 *
	 * @param query a partial object which filter depending on the elements, if not set it will fetch everything
	 * @returns the first element matching with the query or null otherwise
	 */
	public async findOne(query?: Parameters<this['findAll']>[0]): Promise<SchemaInfer<S> | null> {
		return (await this.findAll(query)).data[0] ?? null
	}

	/**
	 * update the remote reference of the object
	 *
	 * note: it will not try to insert an item (use `upsert` to handle this)
	 *
	 * @param obj the object to update
	 * @returns an object if it was able to update or null otherwise
	 */
	public async update(...params: Parameters<NonNullable<DaoAdapter<S>['update']>>): ReturnType<NonNullable<DaoAdapter<S>['update']>> {
		if (!this.adapter.update) {
			throw new Error('the Adapter does not allow you to update to the remote source')
		}
		return this.adapter.update(...params)
	}

	/**
	 * change some elements from the object and return the object updated
	 * @param id the id of the object
	 * @param changes the change to make
	 */
	public async patch(...params: Parameters<NonNullable<DaoAdapter<S>['patch']>>): ReturnType<NonNullable<DaoAdapter<S>['patch']>> {
		if (!this.adapter.patch) {
			const query = await this.findById(params[0])
			if (!query) {
				return null
			}
			return await this.update({ ...query, ...params[1] })
		}
		return await this.adapter.patch(...params)
	}

	/**
	 * update the remote reference of the object or create it if not found
	 * @param obj the object to update/insert
	 * @returns the object is updated/inserted or null otherwise
	 */
	public async upsert(...params: Parameters<NonNullable<DaoAdapter<S>['upsert']>>): ReturnType<NonNullable<DaoAdapter<S>['upsert']>> {
		if (!this.adapter.upsert) {
			throw new Error('the Adapter does not allow you to upsert to the remote source')
		}
		return this.adapter.upsert(...params)
	}

	/**
	 * Delete the object
	 * @param obj the object or ID to delete
	 *
	 * @returns if the object was deleted or not (if object is not in db it will return true)
	 */
	public async delete(...params: Parameters<NonNullable<DaoAdapter<S>['delete']>>): ReturnType<NonNullable<DaoAdapter<S>['delete']>> {
		if (!this.adapter.delete) {
			throw new Error('the Adapter does not allow you to delete on the remote source')
		}
		return this.adapter.delete(...params)
	}
}
